###############################################################################
#
# all.tcl --
#
# This file contains a top-level script to run all of the Garuda tests.
# Execute it by invoking "source all.eagle".
#
# Extensible Adaptable Generalized Logic Engine (Eagle)
# Test Suite File
#
# Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
#
# See the file "license.terms" for information on usage and redistribution of
# this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# RCS: @(#) $Id: $
#
###############################################################################

if {![package vsatisfies [package provide Tcl] 8.4]} then {
  error "need Tcl 8.4 or higher"
}

if {[catch {package present Eagle}] == 0} then {
  error "need native Tcl"
}

namespace eval ::Garuda {
  #############################################################################
  #***************************** SHARED VARIABLES *****************************
  #############################################################################

  #
  # NOTE: Display diagnostic messages while searching for the package being
  #       tested and setting up the tests?  This variable may be shared with
  #       the package being tested; therefore, change it with care.
  #
  variable verbose; # DEFAULT: true

  if {![info exists verbose]} then {
    set verbose true
  }

  #
  # NOTE: The Tcl command used to log warnings, errors, and other messages
  #       generated by the package being tested.  This variable may be shared
  #       with the package being tested; therefore, change it with care.
  #
  variable logCommand; # DEFAULT: tclLog

  if {![info exists logCommand]} then {
    set logCommand tclLog
  }

  #############################################################################
  #****************************** TEST VARIABLES ******************************
  #############################################################################

  #
  # NOTE: Automatically run all the tests now instead of waiting for the
  #       runPackageTests procedure to be executed?
  #
  variable startTests; # DEFAULT: true

  if {![info exists startTests]} then {
    set startTests true
  }

  #
  # NOTE: The build platforms for the package being tested that we know about
  #       and support.
  #
  variable testPackagePlatforms; # DEFAULT: "Win32 x64" OR "x64 Win32"

  if {![info exists testPackagePlatforms]} then {
    #
    # NOTE: Attempt to select the appropriate platforms (architectures)
    #       for this machine.
    #
    if {[info exists ::tcl_platform(machine)] && \
        $::tcl_platform(machine) eq "amd64"} then {
      #
      # NOTE: We are running on an x64 machine, prefer it over x86.
      #
      set testPackagePlatforms [list x64 Win32]
    } else {
      #
      # NOTE: We are running on an x86 machine, prefer it over x64.
      #
      set testPackagePlatforms [list Win32 x64]
    }
  }

  #
  # NOTE: The build configurations for the package being tested that we know
  #       about and support.
  #
  variable testPackageConfigurations; # DEFAULT: {DebugDll ReleaseDll ""}

  if {![info exists testPackageConfigurations]} then {
    set testPackageConfigurations [list DebugDll ReleaseDll ""]
  }

  #
  # NOTE: The name of the package being tested.
  #
  variable testPackageName; # DEFAULT: Garuda

  if {![info exists testPackageName]} then {
    set testPackageName \
        [lindex [split [string trim [namespace current] :] :] 0]
  }

  #
  # NOTE: The version of the package being tested.
  #
  variable testPackageVersion; # DEFAULT: 1.0

  if {![info exists testPackageVersion]} then {
    set testPackageVersion 1.0
  }

  #
  # NOTE: The name of the dynamic link library file containing the native code
  #       for the package being tested.
  #
  variable testBinaryFileName; # DEFAULT: Garuda.dll

  if {![info exists testBinaryFileName]} then {
    set testBinaryFileName $testPackageName[info sharedlibextension]
  }

  #
  # NOTE: The name of the directory where the dynamic link library file
  #       containing the native code for the package being tested resides.
  #
  variable testBinaryPath; # DEFAULT: <unset>

  #
  # NOTE: The names of the Eagle test suite files to run.
  #
  variable testFileNames; # DEFAULT: tcl-load.eagle

  if {![info exists testFileNames]} then {
    set testFileNames [list tcl-load.eagle]
  }

  #
  # NOTE: The name of the main Eagle test suite file.
  #
  variable testSuiteFileName; # DEFAULT: all.eagle

  if {![info exists testSuiteFileName]} then {
    set testSuiteFileName all.eagle
  }

  #############################################################################
  #**************************** SHARED PROCEDURES *****************************
  #############################################################################

  #
  # NOTE: Stolen from "helper.tcl".
  #
  proc maybeFullName { command } {
    set which [namespace which $command]

    if {[string length $which] > 0} then {
      return $which
    }

    return $command
  }

  #
  # NOTE: Stolen from "helper.tcl".
  #
  proc isValidDirectory { path } {
    variable logCommand
    variable verbose

    if {$verbose} then {
      catch {
        set caller [maybeFullName [lindex [info level -1] 0]]

        eval $logCommand [list \
            "$caller: Checking for directory \"$path\"..."]
      }
    }

    #
    # NOTE: For now, just make sure the path refers to an existing directory.
    #
    return [expr {[string length $path] > 0 && [file exists $path] && \
        [file isdirectory $path]}]
  }

  #
  # NOTE: Stolen from "helper.tcl".
  #
  proc isValidFile { path } {
    variable logCommand
    variable verbose

    if {$verbose} then {
      catch {
        set caller [maybeFullName [lindex [info level -1] 0]]

        eval $logCommand [list \
            "$caller: Checking for file \"$path\"..."]
      }
    }

    #
    # NOTE: For now, just make sure the path refers to an existing file.
    #
    return [expr {[string length $path] > 0 && [file exists $path] && \
        [file isfile $path]}]
  }

  #############################################################################
  #**************************** UTILITY PROCEDURES ****************************
  #############################################################################

  proc findPackagePath {
          name version platforms configurations directory fileName } {
    #
    # NOTE: Check the in-development directories for the package being tested,
    #       based on the provided build platforms and configurations.
    #
    foreach platform $platforms {
      foreach configuration $configurations {
        set path [file join $directory bin $platform $configuration \
            $fileName]

        if {[isValidFile $path]} then {
          return [file dirname $path]
        }
      }
    }

    #
    # NOTE: Check the in-deployment directory for the package being tested.
    #
    set path [file join $directory [join [list $name $version] ""] \
        $fileName]

    if {[isValidFile $path]} then {
      return [file dirname $path]
    }

    return ""
  }

  proc addToAutoPath { directory } {
    global auto_path

    if {[lsearch -exact $auto_path $directory] == -1} then {
      lappend auto_path $directory
    }
  }

  #############################################################################
  #************************** TEST STARTUP PROCEDURE **************************
  #############################################################################

  proc runPackageTests { directory } {
    variable envVars
    variable logCommand
    variable rootRegistryKeyName
    variable testBinaryFileName
    variable testBinaryPath
    variable testFileNames
    variable testPackageConfigurations
    variable testPackageName
    variable testPackagePlatforms
    variable testPackageVersion
    variable testSuiteFileName
    variable useEnvironment
    variable useRegistry
    variable verbose

    #
    # NOTE: When running in development [within the source tree], this should
    #       give us the root Eagle source code directory (i.e. the directory
    #       with the "bin", "Library", and "Package" sub-directories).  When
    #       running in deployment (e.g. "<Tcl>\lib\Garuda1.0\tests"), this
    #       should give us the application (or Tcl) library directory (i.e. the
    #       one containing the various package sub-directories).
    #
    set baseDirectory [file dirname [file dirname $directory]]

    #
    # NOTE: Attempt to find binary file for the package being tested using the
    #       configured platforms, configurations, and file name.
    #
    if {[info exists testBinaryPath]} then {
      #
      # NOTE: The path has probably been pre-configured by an external script;
      #       therefore, just use it verbatim.
      #
      set path $testBinaryPath

      if {$verbose} then {
        catch {
          set caller [maybeFullName [lindex [info level 0] 0]]

          eval $logCommand [list \
              "$caller: Using existing binary path \"$testBinaryPath\"..."]
        }
      }
    } else {
      set path [findPackagePath $testPackageName $testPackageVersion \
          $testPackagePlatforms $testPackageConfigurations $baseDirectory \
          $testBinaryFileName]

      if {[isValidDirectory $path]} then {
        set testBinaryPath $path
      }
    }

    #
    # NOTE: Double-check that the configured directory is valid.
    #
    if {[info exists testBinaryPath] && \
        [isValidDirectory $testBinaryPath]} then {
      #
      # NOTE: Success, we found the necessary binary file.  Add the directory
      #       containing the file to the Tcl package search path if it is not
      #       already present.
      #
      if {[lsearch -exact $::auto_path $testBinaryPath] != -1} then {
        if {$verbose} then {
          catch {
            set caller [maybeFullName [lindex [info level 0] 0]]

            eval $logCommand [list \
                "$caller: Binary path already present in \"::auto_path\"..."]
          }
        }
      } else {
        addToAutoPath $testBinaryPath
      }

      #
      # HACK: Scan for and process the "-configuration", "-suffix", "-preTest",
      #       and "-postTest" command line arguments.  The first two are needed
      #       by the "helper.tcl" script to locate the proper Eagle assembly to
      #       load and use for the tests.  The other two are needed to support
      #       various tests.
      #
      foreach {name value} $::argv {
        switch -exact -- $name {
          -configuration -
          -suffix {
            #
            # NOTE: This will be picked up by the "helper.tcl" file.
            #
            set ::test_flags($name) $value

            #
            # NOTE: Show that we set this option (in the log).
            #
            if {$verbose} then {
              catch {
                set caller [maybeFullName [lindex [info level 0] 0]]

                eval $logCommand [list \
                    "$caller: Set test option \"$name\" to value \"$value\"..."]
              }
            }
          }
          -preTest -
          -postTest {
            #
            # NOTE: Set the local variable (minus leading dashes) to the value,
            #       which is a script to evaluate before/after the test itself.
            #
            set [string trimleft $name -] $value

            #
            # NOTE: Show that we set this option (in the log).
            #
            if {$verbose} then {
              catch {
                set caller [maybeFullName [lindex [info level 0] 0]]

                eval $logCommand [list \
                    "$caller: Set test option \"$name\" to value \"$value\"..."]
              }
            }
          }
        }
      }

      #
      # NOTE: Evaluate the pre-test script now, if any.  This must be done
      #       prior to loading the actual Tcl package; otherwise, we cannot
      #       impact the (embedded) Eagle interpreter creation process.
      #
      if {[info exists preTest]} then {
        uplevel #0 $preTest
      }

      #
      # NOTE: Attempt to require the package being tested now.  This should
      #       end up sourcing the "helper.tcl" file, which must also provide
      #       us with the "envVars", "rootRegistryKeyName", "useEnvironment",
      #       and "useRegistry" Tcl variables that we need.
      #
      package require $testPackageName $testPackageVersion

      #
      # NOTE: Configure the Eagle test suite to run only the specified file(s)
      #       unless it has already been configured otherwise.
      #
      if {[lsearch -exact $::argv -file] != -1} then {
        if {$verbose} then {
          catch {
            set caller [maybeFullName [lindex [info level 0] 0]]

            eval $logCommand [list \
                "$caller: Test option \"-file\" already present in \"::argv\"..."]
          }
        }
      } else {
        #
        # NOTE: No file option found, add it.
        #
        lappend ::argv -file $testFileNames

        #
        # NOTE: Show that we set this option (in the log).
        #
        if {$verbose} then {
          catch {
            set caller [maybeFullName [lindex [info level 0] 0]]

            eval $logCommand [list \
                "$caller: Set test option \"-file\" to \"$testFileNames\"..."]
          }
        }
      }

      #
      # NOTE: Build the list of directories to search for the main Eagle test
      #       suite file.
      #
      set testSuiteDirectories [list [file join $baseDirectory Library] \
          $baseDirectory]

      if {$useEnvironment} then {
        eval lappend testSuiteDirectories [getEnvironmentPathList $envVars]
      }

      if {$useRegistry} then {
        eval lappend testSuiteDirectories [getRegistryPathList \
            $rootRegistryKeyName Path]
      }

      if {$verbose} then {
        catch {
          set caller [maybeFullName [lindex [info level 0] 0]]

          eval $logCommand [list \
              "$caller: List of directories to search: $directories"]
        }
      }

      #
      # NOTE: Search for the main Eagle test suite file in all the configured
      #       directories, stopping when found.
      #
      foreach testSuiteDirectory $testSuiteDirectories {
        set testFileName [file join $testSuiteDirectory Tests \
            $testSuiteFileName]

        if {[isValidFile $testFileName]} then {
          break
        }
      }

      #
      # NOTE: Did we find the main Eagle test suite file?
      #
      if {[info exists testFileName] && [isValidFile $testFileName]} then {
        #
        # NOTE: Attempt to run the Eagle test suite now.
        #
        uplevel #0 [list source $testFileName]

        #
        # NOTE: Evaluate the post-test script now, if any.
        #
        if {[info exists postTest]} then {
          uplevel #0 $postTest
        }
      } else {
        error "cannot locate Eagle test suite file: $testSuiteFileName"
      }
    } else {
      error "cannot locate package binary file: $testBinaryFileName"
    }
  }

  #############################################################################
  #******************************* TEST STARTUP *******************************
  #############################################################################

  set packageTestPath [file dirname [info script]]

  if {$startTests} then {
    runPackageTests $packageTestPath
  }
}
