/*
 * Class1.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Components.Public.Delegates;
using Eagle._Constants;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

#if !CONSOLE
using ConsoleColor = Eagle._Components.Public.ConsoleColor;
#endif

namespace Sample
{
    [ObjectId("24642a67-9ef2-4e54-8511-6283669683af")] // FIXME: Always change this.
    internal static class Class1
    {
        [STAThread()] /* WinForms */
        private static int Main(string[] args)
        {
            return (int)Class1.Test(args);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: This represents some state information that Class2 may access
        //       (via the IClientData interface passed to it upon creation).
        //
        internal static int sharedState = 42; // NOTE: Used via Reflection only.

        ///////////////////////////////////////////////////////////////////////////////////////////////

        [MethodFlags(MethodFlags.VariableTrace | MethodFlags.NoAdd)]
        private static ReturnCode TraceCallback(
            BreakpointType breakpointType,
            Interpreter interpreter,
            ITraceInfo traceInfo,
            ref Result result
            )
        {
            //
            // NOTE: Intercept and show all variable access for the interpreter.
            //       We could do practically anything inside this callback,
            //       including monitoring and/or modifying the values to be
            //       fetched and/or set, changing the interpreter state, or
            //       aborting the operation altogether.
            //
            if (interpreter != null)
            {
                if (traceInfo != null)
                {
                    IHost host = interpreter.Host;

                    if (host != null)
                    {
                        try
                        {
                            ConsoleColor foregroundColor = _ConsoleColor.None;
                            ConsoleColor backgroundColor = _ConsoleColor.None;

                            if (host.GetColors(null, "TraceInfo", true, true,
                                    ref foregroundColor, ref backgroundColor,
                                    ref result) == ReturnCode.Ok)
                            {
                                host.Clear();
                                host.ResetPosition();

                                host.WriteTraceInfo(interpreter, traceInfo, true,
                                    true, foregroundColor, backgroundColor);

                                return ReturnCode.Ok;
                            }
                        }
                        catch (Exception e)
                        {
                            result = e;
                        }
                    }
                    else
                    {
                        result = "interpreter host not available";
                    }
                }
                else
                {
                    result = "invalid trace";
                }
            }
            else
            {
                result = "invalid interpreter";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ExitCode Test(string[] args)
        {
            //
            // NOTE: The interpreter exit code (either from exiting the
            //       interactive loop or from an explicit call to the "exit"
            //       command).
            //
            ExitCode exitCode;

            //
            // NOTE: The "interpreter result" that is passed to various
            //       methods.
            //
            Result result = null;

            //
            // NOTE: The error line number that is passed to various methods.
            //
            int errorLine = 0;

            //
            // NOTE: Do we want to intercept all variable access for the
            //       interpreter?
            //
            TraceList traces = new TraceList(new TraceCallback[] {
                TraceCallback
            });

            //
            // NOTE: Setup the creation flags we want for our interpreter.
            //
            CreateFlags flags =
                CreateFlags.Initialize | CreateFlags.ThrowOnDisposed;

            //
            // NOTE: First, we create a new interpreter.
            //
            using (Interpreter interpreter = Interpreter.Create(args, flags, traces, ref result))
            {
                if (interpreter != null)
                {
                    //
                    // NOTE: Create one of our objects to handle our custom
                    //       command.
                    //
                    Class2 class2 = new Class2(new CommandData(
                        "class2", null, null, new ClientData(typeof(Class1)),
                        typeof(Class2).FullName, CommandFlags.None, null, 0));

                    //
                    // NOTE: The token that will represent the custom command
                    //       we add below.
                    //
                    long token = 0;

                    //
                    // NOTE: Next, we optionally add one or more custom
                    //       commands.
                    //
                    /* IGNORED */
                    interpreter.AddCommand(class2, null, ref token, ref result);

                    //
                    // NOTE: Next, evaluate one or more scripts of your 
                    //       choosing (which may or may not reference any
                    //       custom commands you may have added in the previous
                    //       step).
                    //
                    ReturnCode code = Engine.EvaluateScript(
                        interpreter, "class2 test; # <-- script text",
                        ref result, ref errorLine);

                    //
                    // NOTE: Grab the interpreter host now since we need it
                    //       several times below.
                    //
                    IHost host = interpreter.Host;

                    //
                    // NOTE: Handle any script errors.
                    //
                    if (Utility.IsSuccess(code, false))
                    {
                        //
                        // NOTE: Always check for a valid interpreter hosting
                        //       environment before using it as it is not
                        //       guaranteed to always be available.
                        //
                        if (host != null)
                        {
                            host.Clear();

                            host.WriteLine(String.Format(
                                "SUCCESS, code = {0}", code));

                            host.WriteResultLine(code, result);
                        }
                        else
                        {
                            //
                            // NOTE: Do something else here...  Such as log the
                            //       success.
                            //
                        }
                    }
                    else
                    {
                        //
                        // NOTE: Always check for a valid interpreter hosting
                        //       environment before using it as it is not
                        //       guaranteed to always be available.
                        //
                        if (host != null)
                        {
                            host.Clear();

                            host.WriteLine(String.Format(
                                "FAILURE, code = {0}", code));

                            host.WriteResultLine(code, result, errorLine);
                        }
                        else
                        {
                            //
                            // NOTE: Do something else here...  Such as log the
                            //       failure.
                            //
                        }
                    }

                    //
                    // NOTE: Finally, remove any custom commands that may have
                    //       been added.
                    //
                    /* IGNORED */
                    interpreter.RemoveCommand(token, null, ref result);

                    //
                    // NOTE: Fetch the interpreter exit code (typically zero
                    //       for "success" and non-zero for "failure").
                    //
                    exitCode = interpreter.ExitCode;
                }
                else
                {
#if CONSOLE
                    //
                    // NOTE: Creation of the interpreter failed.
                    //
                    Console.WriteLine(Utility.FormatResult(
                        ReturnCode.Error, result));
#endif

                    exitCode = Utility.FailureExitCode();
                }
            }

            return exitCode;
        }
    }
}
